local ADDONNAME, ns = ...

-- load locale
local L = LibStub("AceLocale-3.0"):GetLocale(ADDONNAME)

-- load Libs
local LW = LibStub("LibWho-2.0"):Library()

-- WoW APIs
local _G = _G

-- Lua API
local select, rawget, rawset, type, pairs = select, rawget, rawset, type, pairs
local print, getmetatable, tonumber = print, getmetatable, tonumber

-- GLOBALS: 

ns.lists = {
    whitelist = {},
    blacklist = {},
}

ns.listScopes = {
    global = true,
    profile = true,
}
-- LOCALES: whitelist="Whitelist", blacklist="Blacklist"
ns.listTypes = {
    whitelist = true,
    blacklist = true,
}
-- LOCALES: names="Name", guilds="Guild", realms="Realm"
ns.listEntrys = {
    names = true,
    guilds = true,
    realms = true,
}

local function NameInvite(self, check)
    if check then
        return true
    end
    _G.InviteUnit(self.name)
end

local function BattleNetInvite(self, check)
    if check then
        return true
    end
    local toonID = select(6, _G.BNGetFriendInfoByID(self.presenceID))
    _G.BNInviteFriend(toonID)
end

local function NoopInvite(self, check)
    if check then
        return false
    end
    local why
    if self.isBlacklisted then
        why = L["%s was not invited because %s is blacklisted."]:format(self.name, self.gender)
    elseif self.isNotWhitelisted then
        why = L["%s was not invited because %s is not whitelisted."]:format(self.name, self.gender)
    elseif self.otherGame then
        why = L["%s was not invited because %s does not play World of Warcraft."]:format(self.name, self.gender)
    elseif self.invalidKey then
        why = L["%s was not invited because %s has an invalid ID."]:format(self.name, self.gender)
    elseif self.isNotSameFaction then
        why = L["%s was not invited because %s is not from the same faction."]:format(self.name, self.gender)
    elseif self.invalidRelam then
        why = L["%s was not invited because %s has no realm information."]:format(self.name, self.gender)
    else
        why = L["%s was not invited for an unknown reason."]:format(self.name)
    end

    return why
end

local dash = "-"

local function CheckLists(self, whitelist, blacklist, singelCheck)
    if self:Invite(true) then
        local foundWhitelistEntry = false
        local isWhitelisted = false
        for key, active in pairs(whitelist.names) do
            if active then                
                foundWhitelistEntry = true
                if key:find(dash) then
                    if key:lower() == self.name:lower() then
                        isWhitelisted = true
                        break
                    end
                else
                    if key:lower() == self.cleanName:lower() then
                        isWhitelisted = true
                        break
                    end
                end                
            end
        end
        if not isWhitelisted then
            for key, active in pairs(whitelist.guilds) do
                if active then
                    foundWhitelistEntry = true
                    if key:lower() == self.guild:lower() then
                        isWhitelisted = true
                        break
                    end
                end
            end
        end
        if not isWhitelisted then
            for key, active in pairs(whitelist.realms) do
                if active then
                    foundWhitelistEntry = true
                    if key:lower() == self.realm:lower() then
                        isWhitelisted = true
                        break
                    end
                end
            end
        end
        if foundWhitelistEntry and not isWhitelisted then
            if not singelCheck then
                self.Invite = NoopInvite
                self.isNotWhitelisted = true
            end
            return false, "isNotWhitelisted", NoopInvite
        else
            local isBlacklisted = false
            for key, active in pairs(blacklist.names) do
                if active then
                    if key:find(dash) then
                        if key:lower() == self.name:lower() then
                            isBlacklisted = true
                            break
                        end
                    else
                        if key:lower() == self.cleanName:lower() then
                            isBlacklisted = true
                            break
                        end
                    end                    
                end
            end
            if not isBlacklisted then
                for key, active in pairs(blacklist.guilds) do
                    if active then
                        if key:lower() == self.guild:lower() then
                            isBlacklisted = true
                            break
                        end
                    end
                end
            end
            if not isBlacklisted then
                for key, active in pairs(blacklist.realms) do
                    if active then
                        if key:lower() == self.realm:lower() then
                            isBlacklisted = true
                            break
                        end
                    end
                end
            end
            
            if isBlacklisted then
                if not singelCheck then
                    self.Invite = NoopInvite
                    self.isBlacklisted = true
                end
                return false, "isBlacklisted", NoopInvite
            else
                return true
            end
        end
    else
        return false, nil, NoopInvite
    end
end

local function CheckMainLists(self)
    CheckLists(self, ns.lists.whitelist, ns.lists.blacklist)
end

local function Reset(self)
    if not self.invalidKey then
        self.isBlacklisted = nil
        self.isNotWhitelisted = nil
        self.otherGame = nil
        self.isNotSameFaction = nil
        self.invalidRelam = nil
        self.listsChecked = false

        getmetatable(ns.ID_cache).__index(ns.ID_cache, self.presenceID or self.guid)
    end
end

local LW_opts  = {
    queue = LW.WHOLIB_QUEUE_QUIET,
    timeout = -1,    
}

ns.ID_cache = setmetatable({}, {__index=function(t, key)
    local v = rawget(t, key)

    if type(v) ~= "table" then
        v = {}
    end

    v.CheckLists = CheckLists
    v.CheckMainLists = CheckMainLists
    v.Reset = Reset

    if type(key) == "string" and key:len() == 18 then
        local gender, name, realm = select(5, _G.GetPlayerInfoByGUID(key) )
        if name and realm then            
            v.cleanName = dash:split(name)
            v.realm = realm:len() > 0 and realm or _G.GetRealmName()
            v.name = name:find(dash) and name:gsub("%s*"..dash.."%s*", dash) or name..dash..v.realm
            
            v.guid = key

            v.Invite = NameInvite
            v.gender = gender == 2 and L["he"] or gender == 3 and L["she"] or L["it"]

            LW_opts.callback = function(user, time)
                v.guild = user and user.Guild or ""                
                v:CheckMainLists()
                v.listsChecked = true
            end
            local user = LW:UserInfo(v.cleanName, LW_opts)
            if user then
                v.guild = user.Guild or ""
                v:CheckMainLists()
                v.listsChecked = true
            end               
        else
            v.invalidKey = true
            v.name = name or L["Unknown"]
            v.gender = L["it"]

            v.Invite = NoopInvite
            v.listsChecked = true
            rawset(t, key, v)
            return v
        end
    elseif tonumber(key) then
        local name, client, realm, _, faction, _, _, guild, _, _, _, _, _, _, toonID = select(2, _G.BNGetToonInfo(key) )
        if name then
            v.presenceID = key
            if client ~= _G.BNET_CLIENT_WOW then
                v.otherGame = true
                v.name = name
                v.gender = L["s/he"]               

                v.Invite = NoopInvite
                v.listsChecked = true
                rawset(t, key, v)
                return v
            elseif faction ~= _G.UnitFactionGroup("player") then
                v.isNotSameFaction = true
                v.name = name
                v.gender = L["s/he"]

                v.Invite = NoopInvite
                v.listsChecked = true
                rawset(t, key, v)
                return v
            end

            if realm then
                v.name = name.."-"..realm
                v.gender = L["s/he"]
                v.cleanName = name
                v.realm = realm

                v.toonID = toonID

                v.Invite = BattleNetInvite
                
                LW_opts.callback = function(user, time)
                    v.guild = user and user.Guild or ""
                    v.gender = user and user.Sex == 2 and L["he"] or user and user.Sex == 3 and L["she"] or user and L["it"] or L["s/he"]
                    v:CheckMainLists()
                    v.listsChecked = true
                end
                local user = LW:UserInfo(v.cleanName, LW_opts)
                if user then
                    v.guild = user.Guild or ""
                    v.gender = user.Sex == 2 and L["he"] or user.Sex == 3 and L["she"] or L["it"]
                    v:CheckMainLists()
                    v.listsChecked = true
                end
            else
                v.invalidRelam = true
                v.name = name
                v.gender = L["s/he"]

                v.Invite = NoopInvite
                v.listsChecked = true
                rawset(t, key, v)
                return v
            end
        else
            v.invalidKey = true
            v.name = L["Unknown"]
            v.gender = L["it"]

            v.Invite = NoopInvite
            v.listsChecked = true
            rawset(t, key, v)
            return v
        end
    else
        v.invalidKey = true
        v.name = L["Unknown"]
        v.gender = L["it"]

        v.Invite = NoopInvite
        v.listsChecked = true
        rawset(t, key, v)
        return v
    end        

    rawset(t, key, v)
    return v
end})

-- LOCALES: CHAT_MSG_WHISPER="Whisper", CHAT_MSG_BN_WHISPER="BN Whisper", CHAT_MSG_GUILD="Guild channel", CHAT_MSG_OFFICER="Guild Officer channel", CHAT_MSG_BN_CONVERSATION="BN conversation channel", CHAT_MSG_BN_INLINE_TOAST_BROADCAST="BN Toast"
local defaultChannels = {
    ['*'] = nil,
    CHAT_MSG_WHISPER = true,
    CHAT_MSG_BN_WHISPER = true,
    CHAT_MSG_GUILD = false,
    CHAT_MSG_OFFICER = false,
    CHAT_MSG_BN_CONVERSATION = false,
    CHAT_MSG_BN_INLINE_TOAST_BROADCAST = false,
}

-- build list defaults
local listDefaults = {}
local cacheDefaults = {}
for key in pairs(ns.listEntrys) do
    listDefaults[key] = {}
    cacheDefaults[key] = {}
end

ns.defaults = {
    global = {
        blacklist = listDefaults,
        whitelist = listDefaults,
        cache = cacheDefaults,
    },
    profile = {
        active = true,        
        blacklist = listDefaults,
        whitelist = listDefaults,
        maxInviteScheudles = 10,
        showAdvanceSettings = false,
        keywords = {
            ['**'] = {
                exists = false,
                active = false,
                fullMatch = true,
                capitalization = false,
                showWhyNotInvited = true,
                showScheduleMessage = true,
                autoRaid = true,
                scheduleTime = 1.5,
                channels = defaultChannels,
                blacklist = listDefaults,
                whitelist = listDefaults,
            },
        },        
    },
}